! WHIZARD 1.24 Jan 31 2003
! 
! (C) 1999-2002 by Wolfgang Kilian <kilian@particle.uni-karlsruhe.de>
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module permutations

  implicit none
  private

  type, public :: permutation
     private
     integer, dimension(:), pointer :: p
     integer :: size
  end type permutation
  type, public :: permutation_list
     private
     type(permutation), dimension(:), pointer :: p
     integer :: length, size
  end type permutation_list

  public :: create, destroy
  public :: assignment(=)
  public :: write
  public :: size, len
  public :: component
  public :: permute
  public :: permutation_ok
  public :: transpose
  public :: find_permutation
  public :: make_permutations
  public :: faculty

  interface create
     module procedure permutation_create, permutation_list_create
  end interface
  interface destroy
     module procedure permutation_destroy, permutation_list_destroy
  end interface
  interface assignment (=)
     module procedure permutation_assign
  end interface
  interface write
     module procedure permutation_write_unit, permutation_list_write_unit
  end interface
  interface size
     module procedure permutation_size, permutation_list_size
  end interface
  interface len
     module procedure permutation_list_length
  end interface
  interface component
     module procedure permutation_list_entry
  end interface
  interface transpose
     module procedure permutation_transpose
  end interface

contains

  subroutine permutation_create (p, size)
    type(permutation), intent(inout) :: p
    integer, intent(in) :: size
    integer :: i
    allocate (p%p(size))
    do i=1, size
       p%p(i) = i
    end do
    p%size = size
  end subroutine permutation_create

  subroutine permutation_destroy (p)
    type(permutation), intent(inout) :: p
    deallocate (p%p)
    p%size = 0
  end subroutine permutation_destroy

  subroutine permutation_list_create (p, length, size)
    type(permutation_list), intent(inout) :: p
    integer, intent(in) :: length, size
    integer :: i
    allocate (p%p(length))
    do i=1, length
       call create (p%p(i), size)
    end do
    p%length = length
    p%size = size
  end subroutine permutation_list_create

  subroutine permutation_list_destroy (p)
    type(permutation_list), intent(inout) :: p
    integer :: i
    do i=1, p%length
       call destroy (p%p(i))
    end do
    deallocate (p%p)
    p%length = 0
    p%size = 0
  end subroutine permutation_list_destroy

   subroutine permutation_assign (p, p_in)
    type(permutation), intent(inout) :: p
    type(permutation), intent(in) :: p_in
    integer :: i
    do i=1, p_in%size
       p%p = p_in%p
    end do
  end subroutine permutation_assign

  subroutine permutation_write_unit (u, p)
    integer, intent(in) :: u
    type(permutation), intent (in) :: p
    integer :: i
    do i=1, p%size
       if (p%size<10) then
          write (u,"(1x,I1)", advance="no") p%p(i)
       else
          write (u,"(1x,I2)", advance="no") p%p(i)
       end if
    end do
    write (u, *)
  end subroutine permutation_write_unit

  subroutine permutation_list_write_unit (u, p)
    integer, intent(in) :: u
    type(permutation_list), intent(in) :: p
    integer :: i
    write (u,*) "Permutation list:"
    do i=1, len(p)
       call write (u, p%p(i))
    end do
  end subroutine permutation_list_write_unit

  function permutation_size (perm) result (s)
    type(permutation), intent(in) :: perm
    integer :: s
    s = perm%size
  end function permutation_size

  function permutation_list_size (perm) result (s)
    type(permutation_list), intent(in) :: perm
    integer :: s
    s = perm%size
  end function permutation_list_size

  function permutation_list_length (perm) result (l)
    type(permutation_list), intent(in) :: perm
    integer :: l
    l = perm%length
  end function permutation_list_length

  function permutation_list_entry (pl, i) result (p)
    type(permutation_list), intent(in) :: pl
    integer, intent(in) :: i
    type(permutation) :: p
    if (i>0 .and. i<=len(pl)) then
       p%size = size(pl)
       p%p => pl%p(i)%p
    else
       p%size = 0
       nullify(p%p)
    end if
  end function permutation_list_entry

  function permute (i, p) result (j)
    integer, intent(in) :: i
    type(permutation), intent(in) :: p
    integer :: j
    if (i>0 .and. i<=size(p)) then
       j = p%p(i)
    else
       j = 0
    end if
  end function permute

  function permutation_ok (perm) result (ok)
    type(permutation), intent(in) :: perm
    logical :: ok
    integer :: i
    logical, dimension(perm%size) :: set
    ok = .true.
    set = .false.
    do i=1, perm%size
       ok = (perm%p(i)>0 .and. perm%p(i)<=perm%size)
       if (.not.ok) return
       set(perm%p(i)) = .true.
    end do
    ok = all (set)
  end function permutation_ok

  subroutine permutation_transpose (p, i, j)
    type(permutation), intent(inout) :: p
    integer, intent(in) :: i, j
    integer :: tmp
    tmp = p%p(i)
    p%p(i) = p%p(j)
    p%p(j) = tmp
  end subroutine permutation_transpose

  subroutine find_permutation (perm, a1, a2)
    type(permutation), intent(inout) :: perm
    integer, dimension(:), intent(in) :: a1, a2
    integer :: i, j
    do i=1, size (a1)
       do j=1, size (a2)
          if (a1(i) == a2(j)) then
             perm%p(i) = j
             exit
          end if
          perm%p(i) = 0
       end do
    end do
  end subroutine find_permutation

  subroutine make_permutations (p, code)
    type(permutation_list), intent(inout) :: p
    integer, dimension(:), intent(in) :: code
    logical, dimension(size(code)) :: mask
    logical, dimension(:,:), allocatable :: imask
    integer, dimension(:), allocatable :: n_i
    type(permutation) :: p_init
    integer :: psize, i, j, k, n_different, n
    psize = size (code)
    mask = .true.
    n_different = 0
    do i=1, psize
       if (mask(i)) then
          n_different = n_different + 1
          mask = mask .and. (code /= code(i))
       end if
    end do
    allocate (imask(psize, n_different), n_i(n_different))
    mask = .true.
    k = 0
    do i=1, psize
       if (mask(i)) then
          k = k + 1
          imask(:,k) = (code == code(i))
          n_i(k) = faculty (count(imask(:,k)))
          mask = mask .and. (code /= code(i))
       end if
    end do
    n = product (n_i)
    call create (p, n, psize)
    call create (p_init, psize)
    k = 0
    j = 0
    call perm (p_init)
    call destroy (p_init)
    deallocate (imask, n_i)
  contains
    recursive subroutine perm (p_in)
      type(permutation), intent(in) :: p_in
      type(permutation_list) :: pa
      integer :: i
      k = k + 1
      if (k > n_different) then
         j = j + 1
         p%p(j) = p_in
      else
         call create (pa, n_i(k), psize)
         call make_permutations_s (pa, imask(:,k), p_in)
         do i=1, n_i(k)
            call perm (pa%p(i))
         end do
         call destroy (pa)
      end if
      k = k - 1
    end subroutine perm
  end subroutine make_permutations

  subroutine make_permutations_s (p, mask, p_init)
    type(permutation_list), intent(inout) :: p
    logical, dimension(:), intent(in) :: mask
    type(permutation), intent(in) :: p_init
    type(permutation) :: p_current
    integer :: psize, k
    psize = size (mask)
    call create (p_current, psize)
    k = 0
    call perm (mask, p_init)
    call destroy (p_current)
  contains
    recursive subroutine perm (mask, p_in)
      logical, dimension(:), intent(in) :: mask
      type(permutation), intent(in) :: p_in
      logical, dimension(size(mask)) :: newmask
      integer :: pcount, i, j
      pcount = count (mask)
      select case (pcount)
      case (0)
      case (1)
         k = k + 1
         p%p(k) = p_in
      case default
         i = first (mask)
         newmask = mask
         newmask(i) = .false.
         do j=i, psize
            if (mask(j)) then
               p_current = p_in
               call transpose (p_current, i, j)
               call perm (newmask, p_current)
            end if
         end do
      end select
    end subroutine perm
    function first (mask)
      logical, dimension(:), intent(in) :: mask
      integer :: first, i
      do i=1, size(mask)
         if (mask(i)) then
            first = i
            exit
         end if
      end do
    end function first
  end subroutine make_permutations_s

  function faculty (n) result (f)
    integer, intent(in) :: n
    integer :: f
    integer :: i
    f = 1
    do i=2, abs(n)
       f = f*i
    end do
  end function faculty


end module permutations
